(* #load "unix.cma" *)

(*
Files are processed in order. 
Regions marked appropriately are copied to new files. 
Regions going to the same file need not be countinuous. 
They are appended in order of appearance. 
In addition, line number information is appended in the targer file.
*)

let verbose = ref false;;
let dir = "./code/";;

let string_match r s =
  try let _ = Str.search_forward r s 0 in true
  with Not_found -> false


let copy_prefix = Str.regexp "^%COPYTOFILE=\\([^ \t]*\\)"
let environments = "\\(verbatim\\|program\\|verbatimline\\|lstlisting\\)"
let header_prefix = Str.regexp ("^\\(%\\)")
let begin_prefix = Str.regexp ("^\\\\begin *{" ^ environments ^ "}")
let end_prefix = Str.regexp ("^\\\\end *{" ^ environments ^ "}")

type output = { file : string; outchan : out_channel; name : string }
type input = { infile : string; inchan : in_channel; mutable line : int; }

type openfile = 
    { filename : string; mutable chan : out_channel;
      mutable closed : bool; 
      mutable reused : int };;
let table : (string * openfile) list ref = ref [];;
let size = ref 0

let open_file f =
  (* return an opened chan for f or opens one if needed. 
     firt time opened with Open_trunc, next time reopen with Open_concat *)
  if String.length f = 0 then stdout
  else 
    let path = if f.[0] = '/' then f else dir ^ f in
    try
      let e = List.assoc path !table in
      if e.closed then 
        begin
          e.chan <- open_out_gen [ Open_append ] 0o644 e.filename; 
          e.closed <- false;
          incr size;
          e.chan 
        end
      else e.chan
    with Not_found ->
      (* cleaning *)
      if !size > 10 then
        begin 
          let less (_,x) (_,y) = compare x.closed  y.closed in
          let sorted = List.sort less !table in
          let rec first n l =
            if n > 0 then List.hd l :: first (pred n) (List.tl l) else [] in
          let clean (_,x) = close_out x.chan; x.closed <- true in
          List.iter clean (first 3 sorted);
          size := !size - 3
        end;
      (* creating a new entry *)
      print_string f; print_newline();
      let c = open_out path in
      let e = { filename = path; chan = c; closed = false; reused = 0 } in
      table := (path, e) :: !table; 
      incr size; 
      e.chan;;

let input_line file =
  file.line <- succ file.line; input_line file.inchan;;

let input file  = input_line file

let rec search file =
  let s = input file in
  if string_match copy_prefix s then
    copy_file file (Str.matched_group 1 s)
  else search file

and copy_file file f =
  let insert_header chan h = 
    let l = file.line - List.length h in
    output_string chan
      (Printf.sprintf "#%d \"%s\"\n" l file.infile);
    List.iter
      (fun s -> output_string chan s; output_char chan '\n')
      (List.rev h);
    output_char chan '\n' in
  let rec start header =
    let s =
      try input_line file 
      with End_of_file ->
        let out = open_file f in
        insert_header out header;
        raise End_of_file
    in
    if string_match header_prefix s then
      let n = Str.group_end 1 in
      start (String.sub s n (String.length s - n) :: header)
    else
      let out = open_file f in
      if header = [] then
        output_string out
          (Printf.sprintf "#%d \"%s\"\n" (succ file.line) file.infile)
      else
        insert_header out header;
      if string_match begin_prefix s then
        copy_chan file out  (Str.matched_group 1 s)
      else search file
  in
  start [] 

and copy_chan file out name =
  let s =
    try input_line file
    with End_of_file ->
      failwith 
        (Printf.sprintf "File %s closed with open environment %s"
           file.infile name)
  in
  if string_match end_prefix s then
    let env = Str.matched_group 1 s in
    if env = name then search file 
    else failwith (String.concat " " [ "Env :"; env; name])
  else
    begin
      output_string out s;
      output_char out '\n';
      copy_chan file out name
    end
;;

let main () = 
  let start =
    if Array.length Sys.argv > 2 && Sys.argv.(1) = "-v" then
      (verbose := true; 2)
    else 1 in
  for i = start to Array.length Sys.argv -1
  do
    let file =
      { infile = Sys.argv.(i);
        inchan = open_in Sys.argv.(i);
        line = 0 } in
    if !verbose then
      Printf.fprintf stderr "Processing %s\n" file.infile;
    try search file; close_in file.inchan
    with End_of_file -> 
      close_in file.inchan;
  done;
  List.iter (fun (_,x) -> if not x.closed then close_out x.chan) !table

in main()


